"""
Device handling.
"""


from __future__ import annotations
from typing import Optional
from contextlib import contextmanager
from returnn.tensor import Tensor


__all__ = ["copy_to_device", "get_default_device", "set_default_device", "set_default_device_ctx"]


_default_device: Optional[str] = None


def copy_to_device(x: Tensor, device: Optional[str] = None) -> Tensor:
    """
    Copy tensor to device.

    :param x: tensor
    :param device:
    :return: tensor on device
    """
    if not device:
        device = get_default_device()
    if not device:
        return x
    if x.raw_tensor is None:
        return x
    if x.device == device:
        return x
    # noinspection PyProtectedMember
    return x._raw_backend.copy_to_device(x, device)


def get_default_device() -> Optional[str]:
    """
    :return: default device, where to put new tensors (via random number generators, constant, range_over_dim, etc)
    """
    return _default_device


def set_default_device(device: Optional[str]):
    """
    :param device: see :func:`get_default_device`
    """
    global _default_device
    _default_device = device


@contextmanager
def set_default_device_ctx(device: Optional[str]):
    """
    :param device: see :func:`get_default_device`
    """
    global _default_device
    old_device = _default_device
    try:
        _default_device = device
        yield
    finally:
        _default_device = old_device



#!/usr/bin/env python3

"""
Goes through a dataset (similar as dump-dataset.py), but does the training batch creation,
and calculate statistics about how much seqs we have per batch, how much zero padding there is, etc.
"""

from __future__ import annotations

import os
import sys
import time
import typing
import argparse
import numpy
from pprint import pformat

import _setup_returnn_env  # noqa
import returnn.__main__ as rnn
from returnn.log import log
import returnn.util.basic as util
from returnn.util.basic import Stats, hms, NumbersDict
from returnn.datasets.basic import Batch, Dataset, init_dataset
from returnn.config import Config


config = None  # type: typing.Optional[Config]
dataset = None  # type: typing.Optional[Dataset]


def analyze_dataset(options):
    """
    :param options: argparse.Namespace
    """
    print("Epoch: %i" % options.epoch, file=log.v3)
    print("Dataset keys:", dataset.get_data_keys(), file=log.v3)
    print("Dataset target keys:", dataset.get_target_list(), file=log.v3)
    assert options.key in dataset.get_data_keys()

    terminal_width, _ = util.terminal_size()
    show_interactive_process_bar = log.verbose[3] and (not log.verbose[5]) and terminal_width > 0

    start_time = time.time()
    num_seqs_stats = Stats()
    if options.endseq < 0:
        options.endseq = float("inf")

    recurrent = True
    used_data_keys = dataset.get_data_keys()
    batch_size = config.typed_value("batch_size", 1)
    max_seqs = config.int("max_seqs", -1)
    seq_drop = config.float("seq_drop", 0.0)
    max_seq_length = config.typed_value("max_seq_length", None) or config.float("max_seq_length", 0)
    max_pad_size = config.typed_value("max_pad_size", None)

    batches = dataset.generate_batches(
        recurrent_net=recurrent,
        batch_size=batch_size,
        max_seqs=max_seqs,
        max_seq_length=max_seq_length,
        max_pad_size=max_pad_size,
        seq_drop=seq_drop,
        used_data_keys=used_data_keys,
    )

    step = 0
    total_num_seqs = 0
    total_num_frames = NumbersDict()
    total_num_used_frames = NumbersDict()

    try:
        while batches.has_more():
            # See FeedDictDataProvider.
            (batch,) = batches.peek_next_n(1)
            assert isinstance(batch, Batch)
            if batch.start_seq > options.endseq:
                break
            dataset.load_seqs(batch.start_seq, batch.end_seq)
            complete_frac = batches.completed_frac()
            start_elapsed = time.time() - start_time
            try:
                num_seqs_s = str(dataset.num_seqs)
            except NotImplementedError:
                try:
                    num_seqs_s = "~%i" % dataset.estimated_num_seqs
                except TypeError:  # a number is required, not NoneType
                    num_seqs_s = "?"
            progress_prefix = "%i/%s" % (batch.start_seq, num_seqs_s)
            progress = "%s (%.02f%%)" % (progress_prefix, complete_frac * 100)
            if complete_frac > 0:
                total_time_estimated = start_elapsed / complete_frac
                remaining_estimated = total_time_estimated - start_elapsed
                progress += " (%s)" % hms(remaining_estimated)

            batch_max_time = NumbersDict.max([seq.frame_length for seq in batch.seqs]) * len(batch.seqs)
            batch_num_used_frames = sum([seq.frame_length for seq in batch.seqs], NumbersDict())
            total_num_seqs += len(batch.seqs)
            num_seqs_stats.collect(numpy.array([len(batch.seqs)]))
            total_num_frames += batch_max_time
            total_num_used_frames += batch_num_used_frames

            print(
                "%s, batch %i, num seqs %i, frames %s, used %s (%s)"
                % (
                    progress,
                    step,
                    len(batch.seqs),
                    batch_max_time,
                    batch_num_used_frames,
                    batch_num_used_frames / batch_max_time,
                ),
                file=log.v5,
            )
            if show_interactive_process_bar:
                util.progress_bar_with_time(complete_frac, prefix=progress_prefix)

            step += 1
            batches.advance(1)

    finally:
        print(
            "Done. Total time %s. More seqs which we did not dumped: %s"
            % (hms(time.time() - start_time), batches.has_more()),
            file=log.v2,
        )
        print("Dataset epoch %i, order %r." % (dataset.epoch, dataset.seq_ordering))
        print("Num batches (steps): %i" % step, file=log.v1)
        print("Num seqs: %i" % total_num_seqs, file=log.v1)
        num_seqs_stats.dump(stream=log.v1, stream_prefix="Batch num seqs ")
        for key in used_data_keys:
            print("Data key %r:" % key, file=log.v1)
            print("  Num frames: %s" % total_num_frames[key], file=log.v1)
            print("  Num used frames: %s" % total_num_used_frames[key], file=log.v1)
            print("  Fraction used frames: %s" % (total_num_used_frames / total_num_frames)[key], file=log.v1)
        dataset.finish_epoch()


def init(config_str, config_dataset, use_pretrain, epoch, verbosity):
    """
    :param str config_str: either filename to config-file, or dict for dataset
    :param str|None config_dataset:
    :param bool use_pretrain: might overwrite config options, or even the dataset
    :param int epoch:
    :param int verbosity:
    """
    rnn.init_better_exchook()
    rnn.init_thread_join_hack()
    dataset_opts = None
    config_filename = None
    if config_str.strip().startswith("{"):
        print("Using dataset %s." % config_str)
        dataset_opts = eval(config_str.strip())
    elif config_str.endswith(".hdf"):
        dataset_opts = {"class": "HDFDataset", "files": [config_str]}
        print("Using dataset %r." % dataset_opts)
        assert os.path.exists(config_str)
    else:
        config_filename = config_str
        print("Using config file %r." % config_filename)
        assert os.path.exists(config_filename)
    rnn.init_config(config_filename=config_filename, default_config={"cache_size": "0"})
    global config
    config = rnn.config
    config.set("log", None)
    config.set("log_verbosity", verbosity)
    rnn.init_log()
    print("Returnn %s starting up." % __file__, file=log.v2)
    rnn.returnn_greeting()
    rnn.init_faulthandler()
    util.BackendEngine.select_engine(config=config)
    if not dataset_opts:
        if config_dataset:
            dataset_opts = "config:%s" % config_dataset
        else:
            dataset_opts = "config:train"
    if use_pretrain:
        from returnn.pretrain import pretrain_from_config

        pretrain = pretrain_from_config(config)
        if pretrain:
            print("Using pretrain %s, epoch %i" % (pretrain, epoch), file=log.v2)
            net_dict = pretrain.get_network_json_for_epoch(epoch=epoch)
            if "#config" in net_dict:
                config_overwrites = net_dict["#config"]
                print("Pretrain overwrites these config options:", file=log.v2)
                assert isinstance(config_overwrites, dict)
                for key, value in sorted(config_overwrites.items()):
                    assert isinstance(key, str)
                    orig_value = config.typed_dict.get(key, None)
                    if isinstance(orig_value, dict) and isinstance(value, dict):
                        diff_str = "\n" + util.obj_diff_str(orig_value, value)
                    elif isinstance(value, dict):
                        diff_str = "\n%r ->\n%s" % (orig_value, pformat(value))
                    else:
                        diff_str = " %r -> %r" % (orig_value, value)
                    print("Config key %r for epoch %i:%s" % (key, epoch, diff_str), file=log.v2)
                    config.set(key, value)
            else:
                print("No config overwrites for this epoch.", file=log.v2)
        else:
            print("No pretraining used.", file=log.v2)
    elif config.typed_dict.get("pretrain", None):
        print("Not using pretrain.", file=log.v2)
    dataset_default_opts = {}
    Dataset.kwargs_update_from_config(config, dataset_default_opts)
    print("Using dataset:", dataset_opts, file=log.v2)
    global dataset
    dataset = init_dataset(dataset_opts, default_kwargs=dataset_default_opts)
    assert isinstance(dataset, Dataset)
    dataset.init_seq_order(epoch=epoch)


def main():
    """
    Main entry.
    """
    arg_parser = argparse.ArgumentParser(description="Anaylize dataset batches.")
    arg_parser.add_argument("returnn_config", help="either filename to config-file, or dict for dataset")
    arg_parser.add_argument("--dataset", help="if given the config, specifies the dataset. e.g. 'dev'")
    arg_parser.add_argument("--epoch", type=int, default=1)
    arg_parser.add_argument("--endseq", type=int, default=-1, help="end seq idx (inclusive) or -1 (default: 10)")
    arg_parser.add_argument("--verbosity", type=int, default=5, help="overwrites log_verbosity (default: 4)")
    arg_parser.add_argument("--key", default="data", help="data-key, e.g. 'data' or 'classes'. (default: 'data')")
    arg_parser.add_argument("--use_pretrain", action="store_true")
    args = arg_parser.parse_args()
    init(
        config_str=args.returnn_config,
        config_dataset=args.dataset,
        epoch=args.epoch,
        use_pretrain=args.use_pretrain,
        verbosity=args.verbosity,
    )
    try:
        analyze_dataset(args)
    except KeyboardInterrupt:
        print("KeyboardInterrupt")
        sys.exit(1)
    finally:
        rnn.finalize()


if __name__ == "__main__":
    main()



from __future__ import annotations

import _setup_test_env  # noqa
from subprocess import Popen, PIPE, STDOUT, CalledProcessError
import re
import os
import sys
from glob import glob
import shutil
import unittest
from nose.tools import assert_less, assert_in
from returnn.util import better_exchook
from returnn.util.basic import which_pip


try:
    import torch
except ImportError:
    torch = None
else:
    print("Torch:", torch.__version__)


if "RETURNN_DISABLE_TF" in os.environ and int(os.environ["RETURNN_DISABLE_TF"]) == 1:
    tf = None
else:
    try:
        import tensorflow as tf
    except ImportError:
        tf = None
    else:
        print("TensorFlow:", tf.__version__)


py = sys.executable
print("Python:", py)

os.chdir((os.path.dirname(__file__) or ".") + "/..")
assert os.path.exists("rnn.py")


def build_env(env_update=None):
    """
    :param dict[str,str]|None env_update:
    :return: env dict for Popen
    :rtype: dict[str,str]
    """
    env_update_ = os.environ.copy()
    if env_update:
        env_update_.update(env_update)
    return env_update_


def run(*args, env_update=None, print_stdout=False):
    args = list(args)
    print("run:", args)
    # RETURNN by default outputs on stderr, so just merge both together
    p = Popen(args, stdout=PIPE, stderr=STDOUT, env=build_env(env_update=env_update))
    out, _ = p.communicate()
    out = out.decode("utf8")
    if p.returncode != 0:
        print("Return code is %i" % p.returncode)
        print("std out/err:\n---\n%s\n---\n" % out)
        raise CalledProcessError(cmd=args, returncode=p.returncode, output=out)
    if print_stdout:
        print("std out/err:\n---\n%s\n---\n" % out)
    return out


def parse_last_fer(out: str) -> float:
    """
    :param out:
    :return: FER
    """
    parsed_fer = None
    for line in out.splitlines():
        # example: epoch 5 score: 0.0231807245472 elapsed: 0:00:04 dev: score 0.0137521058997 error 0.00268961807423
        m = re.match("epoch [0-9]+ score: .* dev: .* error ([0-9.]+)\\s?", line)
        if not m:
            # example: dev: score 0.03350000149202181 error 0.009919877954075871
            m = re.match("dev: score .* error ([0-9.]+)\\s?", line)
        if not m:
            continue
        parsed_fer = float(m.group(1))
    err_msg = "ERROR: No epoch dev errors found in output"
    assert parsed_fer is not None, "%s.\nOutput:\n\n%s\n\n%s." % (err_msg, out, err_msg)
    return parsed_fer


def run_and_parse_last_fer(*args, **kwargs):
    out = run(*args, **kwargs)
    return parse_last_fer(out)


def run_config_get_fer(
    config_filename, *args, env_update=None, log_verbosity=5, print_stdout=False, pre_cleanup=True, post_cleanup=True
):
    if pre_cleanup:
        cleanup_tmp_models(config_filename)
    fer = run_and_parse_last_fer(
        py,
        "rnn.py",
        config_filename,
        "++log_verbosity",
        str(log_verbosity),
        *args,
        env_update=env_update,
        print_stdout=print_stdout,
    )
    print("FER:", fer)
    if post_cleanup:
        cleanup_tmp_models(config_filename)
    return fer


def get_model_filename(config_filename: str) -> str:
    assert os.path.exists(config_filename)
    from returnn.config import Config

    config = Config()
    config.load_file(config_filename)
    model_filename = config.value("model", "")
    assert model_filename
    # Remove existing models
    assert model_filename.startswith("/tmp/")
    return model_filename


def cleanup_tmp_models(config_filename: str):
    model_filename = get_model_filename(config_filename)
    for f in glob(model_filename + ".*"):
        os.remove(f)


@unittest.skipIf(not tf, "no TF")
def test_demo_tf_task12ax():
    fer = run_config_get_fer("demos/demo-tf-native-lstm.12ax.config", print_stdout=True)
    # The FER limit here is somewhat arbitrary.
    # It's (more or less) deterministic for some given hardware, some given TF version.
    # Earlier we had limit 0.01, but now that the random order in Task12AXDataset changed,
    # this seems not to be correct anymore, at least in the GitHub CI env.
    # On my local machine (Mac M1), I actually get it quite a bit lower, like 0.00127.
    # I'm not 100% sure that there is maybe sth wrong or not quite optimal...
    assert_less(fer, 0.015)


@unittest.skipIf(not tf, "no TF")
def test_demo_tf_task12ax_eval():
    # The test is specifically for task="eval". We had the problem that the dataset init was broken for this case.
    # Task=eval requires a model though, so first train one epoch.
    cfg_filename = "demos/demo-tf-native-lstm.12ax.config"
    train_dataset_repr = """{"class": "Task12AXDataset", "num_seqs": 10}"""
    dev_dataset_repr = """{"class": "Task12AXDataset", "num_seqs": 10}"""
    fer1 = run_config_get_fer(
        cfg_filename,
        "++num_epochs",
        "2",
        "++train",
        train_dataset_repr,
        "++dev",
        dev_dataset_repr,
        print_stdout=True,
        post_cleanup=False,
    )
    fer2 = run_config_get_fer(
        cfg_filename,
        "++task",
        "eval",
        "++load_epoch",
        "2",
        "++train",
        "None",
        "++dev",
        dev_dataset_repr,
        print_stdout=True,
        pre_cleanup=False,
        post_cleanup=False,
    )
    assert fer1 == fer2
    model_filename = get_model_filename(cfg_filename)
    ep2_files = glob(model_filename + ".002.*")
    assert ep2_files, f"No model files found for epoch 2, {model_filename}"
    for fn in ep2_files:
        shutil.copy(fn, fn.replace(".002.", ".003."))
    fer3 = run_config_get_fer(
        cfg_filename,
        "++task",
        "eval",
        "++load_epoch",
        "3",
        "++train",
        "None",
        "++dev",
        dev_dataset_repr,
        print_stdout=True,
        pre_cleanup=False,
    )
    assert fer3 != fer2  # because random seed, we intentionally did not use fixed_random_seed


@unittest.skipIf(not torch, "no PyTorch")
def test_demo_torch_task12ax():
    cleanup_tmp_models("demos/demo-torch.config")
    out = run(py, "rnn.py", "demos/demo-torch.config", print_stdout=True)
    # Also see test_demo_tf_task12ax above.
    fer = parse_last_fer(out)
    assert_less(fer, 0.02)


def _test_torch_export_to_onnx(cfg_filename: str) -> str:
    """
    Executes the demo passed as a parameter and returns the ONNX exported model as a filename.

    :param cfg_filename: Demo filename, either "demos/demo-rf.config" or "demos/demo-torch.config".
    :return: Filename representing the ONNX model location.
    """
    cleanup_tmp_models(cfg_filename)
    out = run(py, "rnn.py", cfg_filename, "--num_epochs", "1", "--device", "cpu")
    out_pt_model = re.search(r"\nSave model under (.*)\n", out, re.MULTILINE)
    assert out_pt_model, "Could not find the model filename in the output:\n%s" % out
    out_pt_model = out_pt_model.group(1)
    print("*** PT model:", out_pt_model)
    out_onnx_model = out_pt_model.replace(".pt", ".onnx")
    run(py, "tools/torch_export_to_onnx.py", cfg_filename, out_pt_model, out_onnx_model, print_stdout=True)

    return out_onnx_model


def _test_torch_onnx_inference_no_seq_lens(out_onnx_model: str):
    """
    Tests the inference of the torch demo with an ONNX model passed as parameter.
    """
    import onnxruntime as ort

    torch.manual_seed(42)
    # Set arbitrary values to pass to the model, respecting the input size (9).
    dummy_data = torch.randn([3, 50, 9])

    session = ort.InferenceSession(out_onnx_model)
    outputs_onnx = session.run(
        None,
        {
            "data": dummy_data.numpy(),
        },
    )
    print("*** Result:", torch.FloatTensor(outputs_onnx[0]))
    print("*** Batch size:", torch.IntTensor(outputs_onnx[1]))
    print("*** Sequence lengths:", torch.IntTensor(outputs_onnx[2]))


def _test_torch_onnx_inference_seq_lens_in_out(out_onnx_model: str):
    """
    Tests the inference of the torch demo with an ONNX model passed as parameter.
    """
    print(out_onnx_model)
    import onnxruntime as ort

    torch.manual_seed(42)
    # Set arbitrary values to pass to the model, respecting the input size (9).
    dummy_data = torch.randn([3, 50, 9])
    dummy_seq_lens = torch.tensor([27, 50, 43], dtype=torch.int32)

    session = ort.InferenceSession(out_onnx_model)
    outputs_onnx = session.run(
        None,
        {
            "data": dummy_data.numpy(),
            "data:size1": dummy_seq_lens.numpy(),
        },
    )
    out_result = torch.FloatTensor(outputs_onnx[0])
    out_seq_lens = torch.IntTensor(outputs_onnx[1])
    print("*** Result:", out_result)
    print("*** Sequence lengths:", out_seq_lens)

    assert out_result.shape == torch.Size([3, 50, 2])


@unittest.skipIf(not torch, "no PyTorch")
def test_demo_torch_export_to_onnx():
    out_onnx_model = _test_torch_export_to_onnx("demos/demo-torch.config")
    _test_torch_onnx_inference_seq_lens_in_out(out_onnx_model)


@unittest.skipIf(not torch, "no PyTorch")
def test_demo_rf_export_to_onnx():
    out_onnx_model = _test_torch_export_to_onnx("demos/demo-rf.config")
    _test_torch_onnx_inference_seq_lens_in_out(out_onnx_model)


@unittest.skipIf(not torch, "no PyTorch")
def test_demo_rf_torch_task12ax():
    cleanup_tmp_models("demos/demo-rf.config")
    out = run(py, "rnn.py", "demos/demo-rf.config", print_stdout=True)
    # Also see test_demo_tf_task12ax above.
    fer = parse_last_fer(out)
    assert_less(fer, 0.02)


@unittest.skipIf(not tf, "no TF")
def test_demo_rf_tf_task12ax():
    cleanup_tmp_models("demos/demo-rf.config")
    out = run(py, "rnn.py", "demos/demo-rf.config", "++backend", "tensorflow-net-dict", print_stdout=True)
    # Also see test_demo_tf_task12ax above.
    fer = parse_last_fer(out)
    assert_less(fer, 0.02)


def test_demo_iter_dataset_task12ax():
    # there should be no actual TF dependency, we just iterate the dataset
    cleanup_tmp_models("demos/demo-tf-vanilla-lstm.12ax.config")
    # pick any 12ax config for the dataset test
    out = run(py, "demos/demo-iter-dataset.py", "demos/demo-tf-vanilla-lstm.12ax.config")
    assert_in("Epoch 5.", out.splitlines())


@unittest.skipIf(not tf, "no TF")
def test_demo_returnn_as_framework():
    print("Prepare.")
    import subprocess
    import shutil
    from glob import glob
    from returnn.util.basic import get_login_username

    # echo via subprocess, because this stdout as well as the other will always be visible.
    subprocess.check_call(["echo", "travis_fold:start:test_demo_returnn_as_framework"])
    assert os.path.exists("setup.py")
    if glob("dist/*.tar.gz"):
        # we want it unique below
        for fn in glob("dist/*.tar.gz"):
            os.remove(fn)
    if os.path.exists("MANIFEST"):
        os.remove("MANIFEST")  # auto-generated. will be recreated
    if os.path.exists("docs/crnn"):
        os.remove("docs/crnn")  # this is auto-generated, and confuses setup.py sdist
    if os.path.exists("docs/returnn"):
        os.remove("docs/returnn")  # this is auto-generated, and confuses setup.py sdist
    tmp_model_dir = "/tmp/%s/returnn-demo-as-framework" % get_login_username()
    if os.path.exists(tmp_model_dir):
        shutil.rmtree(tmp_model_dir, ignore_errors=True)
    print("setup.py sdist, to create package.")
    subprocess.check_call([py, "setup.py", "sdist"])
    dist_fns = glob("dist/*.tar.gz")
    assert len(dist_fns) == 1
    dist_fn = os.path.abspath(dist_fns[0])
    pip_path = which_pip()
    print("Pip install Returnn.")
    in_virtual_env = hasattr(sys, "real_prefix")  # https://stackoverflow.com/questions/1871549/
    cmd = [py, pip_path, "install"]
    if not in_virtual_env:
        cmd += ["--user"]
    cmd += ["-v", dist_fn]
    print("$ %s" % " ".join(cmd))
    subprocess.check_call(cmd, cwd="/")
    print("Running demo now.")
    subprocess.check_call([py, "demo-returnn-as-framework.py"], cwd="demos")
    print("Success.")
    subprocess.check_call(["echo", "travis_fold:end:test_demo_returnn_as_framework"])


@unittest.skipIf(not tf, "no TF")
def test_demo_sprint_interface():
    import subprocess

    # echo via subprocess, because this stdout as well as the other will always be visible.
    subprocess.check_call(["echo", "travis_fold:start:test_demo_sprint_interface"])
    subprocess.check_call([py, os.path.abspath("demos/demo-sprint-interface.py")], cwd="/")
    subprocess.check_call(["echo", "travis_fold:end:test_demo_sprint_interface"])


def test_returnn_as_framework_TaskSystem():
    import subprocess

    # echo via subprocess, because this stdout as well as the other will always be visible.
    subprocess.check_call(["echo", "travis_fold:start:test_returnn_as_framework_TaskSystem"])
    subprocess.check_call([py, os.path.abspath("tests/returnn-as-framework.py"), "test_TaskSystem_Pickler()"], cwd="/")
    subprocess.check_call(["echo", "travis_fold:end:test_returnn_as_framework_TaskSystem"])


@unittest.skipIf(not tf, "no TF")
def test_returnn_as_framework_old_style_crnn_TFUtil():
    """
    Check that old-style `import crnn.TFUtil` works.

    It's not so much about TFUtil, it also could be some other module.
    It's about the old-style module names.
    This is the logic in __old_mod_loader__.
    """
    import subprocess

    # echo via subprocess, because this stdout as well as the other will always be visible.
    subprocess.check_call(["echo", "travis_fold:start:test_returnn_as_framework_old_style_crnn_TFUtil"])
    subprocess.check_call(
        [
            py,
            os.path.abspath("tests/returnn-as-framework.py"),
            "--old-style",
            "--returnn-package-name",
            "crnn",
            "test_old_style_import_crnn_TFUtil()",
        ],
        cwd="/",
    )
    subprocess.check_call(["echo", "travis_fold:end:test_returnn_as_framework_old_style_crnn_TFUtil"])


@unittest.skipIf(not tf, "no TF")
def test_returnn_as_framework_old_style_TFUtil():
    """
    Check that old-style `import TFUtil` works.
    See also :func:`test_returnn_as_framework_old_style_crnn_TFUtil`.
    """
    import subprocess

    # echo via subprocess, because this stdout as well as the other will always be visible.
    subprocess.check_call(["echo", "travis_fold:start:test_returnn_as_framework_old_style_TFUtil"])
    subprocess.check_call(
        [py, os.path.abspath("tests/returnn-as-framework.py"), "--old-style", "test_old_style_import_TFUtil()"], cwd="/"
    )
    subprocess.check_call(["echo", "travis_fold:end:test_returnn_as_framework_old_style_TFUtil"])


if __name__ == "__main__":
    better_exchook.install()
    import sys

    fns = sys.argv[1:]
    if not fns:
        fns = [arg for arg in globals() if arg.startswith("test_")]
    for arg in fns:
        f = globals()[arg]
        print("-" * 20)
        print("Run:", f)
        print("-" * 20)
        try:
            f()
        except unittest.SkipTest as exc:
            print("(SkipTest: %s)" % exc)
        print("-" * 20)
        print("Ok.")
        print("-" * 20)



"""
Conformer model, variant of Transformer with additional convolution, introduced for speech recognition.
Ref: https://arxiv.org/abs/2005.08100

About details of the specific implementation and other implementations, see:
https://github.com/rwth-i6/returnn_common/issues/233
"""

from __future__ import annotations
from typing import Optional, Union, Any, Tuple, List, Dict, Callable
import copy as _copy
from returnn.tensor import Tensor, Dim
import returnn.frontend as rf
from returnn.util.basic import NotSpecified
from .base import ISeqDownsamplingEncoder
from ..decoder.transformer import FeedForward, make_norm


class ConformerPositionwiseFeedForward(FeedForward):
    """
    Conformer position-wise feedforward neural network layer
        FF -> Activation -> Dropout -> FF
    """

    def __init__(
        self,
        out_dim: Dim,
        *,
        ff_dim: Union[Dim, int] = NotSpecified,
        dropout: float = 0.1,
        activation: Union[Callable[[Tensor], Tensor], Dict[str, Any], rf.Module] = rf.swish,
        **kwargs,
    ):
        """
        :param out_dim: output feature dimension
        :param ff_dim: dimension of the feed-forward layers
        :param dropout: dropout value
        :param activation: activation function. swish by default, unlike the base :class:`FeedForward`
        """
        if activation is NotSpecified:
            activation = rf.swish
        super().__init__(out_dim=out_dim, ff_dim=ff_dim, dropout=dropout, activation=activation, **kwargs)


class ConformerConvBlock(rf.Module):
    """
    Conformer convolution block
        FF -> GLU -> depthwise conv -> BN -> Swish -> FF
    """

    def __init__(self, out_dim: Dim, *, kernel_size: int, norm: Union[rf.BatchNorm, Any]):
        """
        :param out_dim: output feature dimension
        :param kernel_size: kernel size of depthwise convolution
        :param norm: Batch norm originally
        """
        super().__init__()
        self.out_dim = out_dim

        self.positionwise_conv1 = rf.Linear(out_dim, 2 * out_dim)
        self.depthwise_conv = rf.Conv1d(
            out_dim, out_dim, filter_size=kernel_size, groups=out_dim.dimension, padding="same"
        )
        self.positionwise_conv2 = rf.Linear(out_dim, out_dim)
        if not callable(norm):
            raise TypeError(f"{self}: unexpected norm type {norm!r}")
        self.norm = norm

    def __call__(self, inp: Tensor, *, spatial_dim: Dim) -> Tensor:
        """forward"""
        x_conv1 = self.positionwise_conv1(inp)
        x_act, _ = rf.gating(x_conv1)
        x_depthwise_conv, _ = self.depthwise_conv(x_act, in_spatial_dim=spatial_dim)
        x_normed = self.norm(x_depthwise_conv)
        x_swish = rf.swish(x_normed)
        x_conv2 = self.positionwise_conv2(x_swish)
        return x_conv2


class ConformerConvSubsample(ISeqDownsamplingEncoder):
    """
    Conv 2D block with optional max-pooling or striding.

    References:

      https://github.com/espnet/espnet/blob/4138010fb66ad27a43e8bee48a4932829a0847ae/espnet/nets/pytorch_backend/transformer/subsampling.py#L162
      https://github.com/rwth-i6/returnn-experiments/blob/5852e21f44d5450909dee29d89020f1b8d36aa68/2022-swb-conformer-hybrid-sat/table_1_and_2/reduced_dim.config#L226
      (actually the latter is different...)

    To get the ESPnet case, for example Conv2dSubsampling6, use these options
    (out_dim is the model dim of the encoder)

      out_dims=[out_dim, out_dim],  # ESPnet standard, but this might be too large
      filter_sizes=[3, 5],
      strides=[2, 3],
      padding="valid",
    """

    def __init__(
        self,
        in_dim: Dim,
        *,
        out_dims: List[Dim],
        filter_sizes: List[Union[int, Tuple[int, int]]],
        strides: Optional[List[Union[int, Tuple[int, int]]]] = None,
        pool_sizes: Optional[List[Tuple[int, int]]] = None,
        activation: Callable[[Tensor], Tensor] = rf.relu,
        padding: str = "same",
    ):
        """
        :param out_dims: the number of output channels. last element is the output feature dimension
        :param filter_sizes: a list of filter sizes for the conv layer
        :param pool_sizes: a list of pooling factors applied after conv layer
        :param activation: the activation function
        :param padding: 'same' or 'valid'
        """
        super().__init__()

        self.pool_sizes = pool_sizes
        self.activation = activation

        self.conv_layers: rf.ModuleList[rf.Conv2d] = rf.ModuleList()
        if strides is None:
            strides = [1] * len(out_dims)
        assert len(out_dims) == len(filter_sizes) == len(strides) > 0
        self._dummy_in_dim = Dim(1, name="dummy-input-feature-dim")
        self.in_dim = in_dim
        prev_out_dim = self._dummy_in_dim
        second_spatial_dim = in_dim
        for i, (filter_size, stride, out_dim) in enumerate(zip(filter_sizes, strides, out_dims)):
            conv = rf.Conv2d(prev_out_dim, out_dim, filter_size=filter_size, strides=stride, padding=padding)
            self.conv_layers.append(conv)
            (second_spatial_dim,) = rf.make_conv_out_spatial_dims(
                [second_spatial_dim], filter_size=conv.filter_size[1], strides=conv.strides[1], padding=padding
            )
            if self.pool_sizes and i < len(self.pool_sizes):
                (second_spatial_dim,) = rf.make_conv_out_spatial_dims(
                    [second_spatial_dim],
                    filter_size=self.pool_sizes[i][1],
                    strides=self.pool_sizes[i][1],
                    padding="same",
                )
            prev_out_dim = out_dim
        self._final_second_spatial_dim = second_spatial_dim
        self.out_dim = second_spatial_dim * prev_out_dim

    def __call__(self, source: Tensor, *, in_spatial_dim: Dim) -> Tuple[Tensor, Dim]:
        """forward"""
        assert self.in_dim in source.dims
        in_spatial_dims = [in_spatial_dim, self.in_dim]
        in_dim = self._dummy_in_dim
        x = rf.expand_dim(source, dim=in_dim)
        for i, conv_layer in enumerate(self.conv_layers):
            x, in_spatial_dims = conv_layer(x, in_spatial_dims=in_spatial_dims)
            in_dim = conv_layer.out_dim
            x = self.activation(x)
            if self.pool_sizes and i < len(self.pool_sizes):
                x, in_spatial_dims = rf.pool2d(
                    x, in_spatial_dims=in_spatial_dims, pool_size=self.pool_sizes[i], padding="same", mode="max"
                )
        x, in_spatial_dims[-1] = rf.replace_dim(x, out_dim=self._final_second_spatial_dim, in_dim=in_spatial_dims[-1])
        out, _ = rf.merge_dims(x, dims=[self._final_second_spatial_dim, in_dim])
        return out, in_spatial_dims[0]


class ConformerEncoderLayer(rf.Module):
    """
    Represents a conformer block
    """

    def __init__(
        self,
        out_dim: Dim = Dim(512, name="conformer-enc-default-out-dim"),
        *,
        ff: Union[type, Dict[str, Any], rf.Module] = NotSpecified,
        ff_dim: Dim = NotSpecified,
        ff_activation: Union[Callable[[Tensor], Tensor], Dict[str, Any], rf.Module] = NotSpecified,
        dropout: float = 0.1,
        conv_kernel_size: int = 32,
        conv_norm: Union[rf.BatchNorm, type, Dict[str, Any], Any] = NotSpecified,
        conv_norm_opts: Optional[Dict[str, Any]] = None,
        num_heads: int = 4,
        self_att: Optional[Union[rf.RelPosSelfAttention, rf.Module, type, Dict[str, Any], Any]] = None,
        self_att_opts: Optional[Dict[str, Any]] = None,
        att_dropout: float = 0.1,
        norm: Union[type, Dict[str, Any], rf.Module, Callable] = rf.LayerNorm,
    ):
        """
        :param out_dim: the output feature dimension
        :param ff_dim: the dimension of feed-forward layers. 2048 originally, or 4 times out_dim
        :param ff_activation: activation function for feed-forward network
        :param dropout: the dropout value for the FF block
        :param conv_kernel_size: the kernel size of depthwise convolution in the conv block
        :param conv_norm: used for the conv block. Batch norm originally
        :param conv_norm_opts: for nn.BatchNorm or other conv_norm type.
          In case of nn.BatchNorm, uses use_mask=False by default.
            use_mask means whether to properly mask the spatial dim in batch norm.
            Most existing implementations don't do this. Except of RETURNN.
            It's faster when you don't do this.
        :param num_heads: the number of attention heads
        :param self_att: the self-attention layer. RelPosSelfAttention originally and default
        :param self_att_opts: options for the self-attention layer, for :class:`nn.RelPosSelfAttention`
        :param att_dropout: attention dropout value
        :param norm: pre-normalization for FF, conv and attention blocks
        """
        super().__init__()

        self.dropout = dropout
        self.dropout_broadcast = rf.dropout_broadcast_default()
        self.out_dim = out_dim

        self.ffn1 = make_ff(ff=ff, out_dim=out_dim, ff_dim=ff_dim, dropout=dropout, ff_activation=ff_activation)
        self.ffn1_layer_norm = make_norm(norm, out_dim)

        self.ffn2 = make_ff(ff=ff, out_dim=out_dim, ff_dim=ff_dim, dropout=dropout, ff_activation=ff_activation)
        self.ffn2_layer_norm = make_norm(norm, out_dim)

        if conv_norm is NotSpecified or conv_norm is rf.BatchNorm:
            conv_norm_opts = conv_norm_opts.copy() if conv_norm_opts else {}
            conv_norm_opts.setdefault("use_mask", False)
            conv_norm = rf.BatchNorm(out_dim, **conv_norm_opts)
        elif isinstance(conv_norm, type):
            conv_norm = conv_norm(out_dim, **(conv_norm_opts or {}))
        elif isinstance(conv_norm, dict):
            conv_norm = rf.build_from_dict(conv_norm, out_dim, **(conv_norm_opts or {}))
        if not callable(conv_norm):
            raise TypeError(f"{self}: unexpected conv_norm type {conv_norm!r}")
        self.conv_block = ConformerConvBlock(out_dim=out_dim, kernel_size=conv_kernel_size, norm=conv_norm)
        self.conv_layer_norm = make_norm(norm, out_dim)

        if self_att is None or isinstance(self_att, (dict, type)):
            self_att_opts_ = dict(
                in_dim=out_dim,
                proj_dim=out_dim,
                key_dim_total=out_dim,
                value_dim_total=out_dim,
                num_heads=num_heads,
                att_dropout=att_dropout,
            )
            if self_att_opts:
                self_att_opts_.update(self_att_opts)
            if self_att is None:
                self.self_att = rf.RelPosSelfAttention(**self_att_opts_)
            elif isinstance(self_att, type):
                self.self_att = self_att(**self_att_opts_)
            elif isinstance(self_att, dict):
                self_att_opts_ = {k: v for (k, v) in self_att_opts_.items() if k not in self_att}
                self.self_att = rf.build_from_dict(self_att, **self_att_opts_)
            else:
                raise TypeError(f"{self}: invalid type: self_att {self_att!r}")
        else:
            if not callable(self_att):
                raise TypeError(f"{self}: invalid non-callable: self_att {self_att!r}")
            self.self_att = self_att
        self.self_att_layer_norm = make_norm(norm, out_dim)

        self.final_layer_norm = make_norm(norm, out_dim)

    def __call__(self, inp: Tensor, *, spatial_dim: Dim) -> Tensor:
        """forward"""
        # FFN
        x_ffn1_ln = self.ffn1_layer_norm(inp)
        x_ffn1 = self.ffn1(x_ffn1_ln)
        x_ffn1_out = 0.5 * rf.dropout(x_ffn1, self.dropout, axis=self.dropout_broadcast and self.out_dim) + inp

        # MHSA
        x_mhsa_ln = self.self_att_layer_norm(x_ffn1_out)
        x_mhsa = self.self_att(x_mhsa_ln, axis=spatial_dim)
        x_mhsa = rf.dropout(x_mhsa, self.dropout, axis=self.dropout_broadcast and self.out_dim)
        x_mhsa_out = x_mhsa + x_ffn1_out

        # Conv
        x_conv_ln = self.conv_layer_norm(x_mhsa_out)
        x_conv = self.conv_block(x_conv_ln, spatial_dim=spatial_dim)
        x_conv_out = rf.dropout(x_conv, self.dropout, axis=self.dropout_broadcast and self.out_dim) + x_mhsa_out

        # FFN
        x_ffn2_ln = self.ffn2_layer_norm(x_conv_out)
        x_ffn2 = self.ffn2(x_ffn2_ln)
        x_ffn2_out = 0.5 * rf.dropout(x_ffn2, self.dropout, axis=self.dropout_broadcast and self.out_dim) + x_conv_out

        # last LN layer
        return self.final_layer_norm(x_ffn2_out)


class ConformerEncoder(ISeqDownsamplingEncoder):
    """
    Represents Conformer encoder architecture
    """

    def __init__(
        self,
        in_dim: Dim,
        out_dim: Dim = Dim(512, name="conformer-enc-default-out-dim"),
        *,
        num_layers: int,
        input_layer: Optional[Union[ConformerConvSubsample, ISeqDownsamplingEncoder, rf.Module, Any]],
        input_embedding_scale: float = 1.0,
        input_dropout: float = 0.1,
        ff_dim: Dim = NotSpecified,
        ff_activation: Union[Callable[[Tensor], Tensor], Dict[str, Any], rf.Module] = NotSpecified,
        dropout: float = 0.1,
        conv_kernel_size: int = NotSpecified,
        conv_norm: Union[rf.BatchNorm, type, Dict[str, Any], Any] = NotSpecified,
        num_heads: int = 4,
        att_dropout: float = 0.1,
        encoder_layer: Optional[Union[ConformerEncoderLayer, rf.Module, type, Dict[str, Any], Any]] = None,
        encoder_layer_opts: Optional[Dict[str, Any]] = None,
        sequential=rf.Sequential,
    ):
        """
        :param out_dim: the output feature dimension
        :param num_layers: the number of encoder layers
        :param input_layer: input/frontend/prenet with potential subsampling.
            (x, in_spatial_dim) -> (y, out_spatial_dim)
        :param input_embedding_scale: applied after input_layer. 1.0 by default for historic reasons.
            In std Transformer, also ESPnet E-Branchformer and Conformer, this is sqrt(out_dim).
        :param input_dropout: applied after input_projection(input_layer(x))
        :param ff_dim: the dimension of feed-forward layers. 2048 originally, or 4 times out_dim
        :param ff_activation: activation function for feed-forward network
        :param dropout: the dropout value for the FF block
        :param conv_kernel_size: the kernel size of depthwise convolution in the conv block
        :param conv_norm: used for the conv block. Batch norm originally
        :param num_heads: the number of attention heads
        :param att_dropout: attention dropout value
        :param encoder_layer: an instance of :class:`ConformerEncoderLayer` or similar
        :param encoder_layer_opts: options for the encoder layer
        :param sequential:
        """
        super().__init__()

        self.in_dim = in_dim
        self.out_dim = out_dim
        self.dropout = dropout
        self.dropout_broadcast = rf.dropout_broadcast_default()

        # TODO once we figured out good defaults, we would create ConformerConvSubsample here when not given
        self.input_layer = input_layer
        self.input_projection = (
            rf.Linear(self.input_layer.out_dim if self.input_layer else self.in_dim, self.out_dim, with_bias=False)
            if input_layer
            else None
        )
        self.input_embedding_scale = input_embedding_scale
        self.input_dropout = input_dropout

        if not encoder_layer or isinstance(encoder_layer, (dict, type)):
            encoder_layer_opts_ = dict(
                out_dim=out_dim,
                ff_dim=ff_dim,
                ff_activation=ff_activation,
                dropout=dropout,
                conv_kernel_size=conv_kernel_size,
                conv_norm=conv_norm,
                num_heads=num_heads,
                att_dropout=att_dropout,
            )
            encoder_layer_opts_ = {k: v for (k, v) in encoder_layer_opts_.items() if v is not NotSpecified}
            if encoder_layer_opts:
                encoder_layer_opts_.update(encoder_layer_opts)
            if not encoder_layer:
                encoder_layer = ConformerEncoderLayer(**encoder_layer_opts_)
            elif isinstance(encoder_layer, type):
                encoder_layer = encoder_layer(**encoder_layer_opts_)
            elif isinstance(encoder_layer, dict):
                # Note: Reuse all the encoder_layer_opts_.
                # If this does not make sense for the specific encoder_layer class here,
                # we would suggest to use a different ConformerEncoder class.
                # (The alternative, to not reuse encoder_layer_opts_ here,
                #  would probably be more confusing, as those options are all ignored then.
                #  It's also not clear what args to pass then and what not.)
                # (Maybe we should do a ConformerEncoderV2 if this is confusing here...)
                encoder_layer_opts_ = {k: v for (k, v) in encoder_layer_opts_.items() if k not in encoder_layer}
                encoder_layer = rf.build_from_dict(encoder_layer, **encoder_layer_opts_)
            else:
                raise TypeError(f"unexpected encoder_layer {encoder_layer!r}")
        else:
            if not callable(encoder_layer):
                raise TypeError(f"{self}: invalid non-callable encoder_layer {encoder_layer!r}")

        self.layers = sequential(_copy.deepcopy(encoder_layer) for _ in range(num_layers))

    def __call__(
        self,
        source: Tensor,
        *,
        in_spatial_dim: Dim,
        collected_outputs: Optional[Dict[str, Tensor]] = None,
    ) -> Tuple[Tensor, Dim]:
        """forward"""
        if self.input_layer:
            x_subsample, out_spatial_dim = self.input_layer(source, in_spatial_dim=in_spatial_dim)
        else:
            x_subsample, out_spatial_dim = source, in_spatial_dim
        x = self.input_projection(x_subsample) if self.input_projection else x_subsample
        if self.input_embedding_scale != 1.0:
            x = x * self.input_embedding_scale
        x = rf.dropout(x, self.input_dropout, axis=self.dropout_broadcast and self.out_dim)
        x = self.layers(x, spatial_dim=out_spatial_dim, collected_outputs=collected_outputs)
        return x, out_spatial_dim


def make_ff(
    *,
    out_dim: Dim,
    ff: Union[type, Dict[str, Any], rf.Module],
    ff_dim: Union[Dim, int],
    ff_activation: Union[Callable[[Tensor], Tensor], Dict[str, Any], rf.Module],
    dropout: float,
) -> Union[ConformerPositionwiseFeedForward, rf.Module]:
    """
    make the feed-forward part of the Conformer layer
    """
    if ff is NotSpecified:
        ff = ConformerPositionwiseFeedForward
    if isinstance(ff, rf.Module):
        ff = _copy.deepcopy(ff)
    else:
        ff_kwargs = dict(out_dim=out_dim, ff_dim=ff_dim, dropout=dropout, activation=ff_activation)
        ff_kwargs = {k: v for (k, v) in ff_kwargs.items() if v is not NotSpecified}
        if isinstance(ff, type):
            ff = ff(**ff_kwargs)
        elif isinstance(ff, dict):
            ff = rf.build_from_dict(ff, **ff_kwargs)
        else:
            raise TypeError(f"unexpected ff type {ff!r}")
    assert isinstance(ff, rf.Module)
    return ff



#!/usr/bin/env python3

"""
This is mostly intended to be run as a test.
This also demonstrates how the SprintInterface is being used by Sprint (RASR).
"""

import os
import tempfile
import sys


_my_dir = os.path.dirname(os.path.abspath(__file__))
_base_dir = os.path.dirname(_my_dir)
assert os.path.exists("%s/rnn.py" % _base_dir)


def main():
    """
    Main entry.
    """
    tmp_dir = tempfile.mkdtemp()
    os.symlink("%s/returnn" % _base_dir, "%s/returnn" % tmp_dir)
    config_fn = "%s/returnn.config" % tmp_dir
    with open(config_fn, "w") as f:
        f.write("#!rnn.py\n")  # Python format
        f.write("use_tensorflow = True\n")
        f.write("num_inputs, num_outputs = 3, 5\n")
        f.write("network = {'output': {'class': 'softmax', 'target': 'classes'}}\n")
        f.write("model = %r + '/model'\n" % tmp_dir)
    open("%s/model.001.meta" % tmp_dir, "w").close()
    sys.path.insert(0, tmp_dir)
    print("Import SprintInterface (relative import).")
    import returnn.sprint.interface

    print("SprintInterface.init")
    returnn.sprint.interface.init(
        inputDim=3,
        outputDim=5,
        cudaEnabled=0,
        targetMode="forward-only",
        config="epoch:1,action:nop,configfile:%s" % config_fn,
    )  # normally action:forward
    print("Ok.")


if __name__ == "__main__":
    main()
    print("demo-sprint-interface exit.")



"""
Types
"""

from __future__ import annotations
from typing import Union, Type, TYPE_CHECKING, Sequence
import numpy
from returnn.tensor import Tensor, Dim, TensorDict
import returnn.frontend as rf


try:
    from typing import Protocol
except ImportError:
    try:
        from typing_extensions import Protocol
    except ImportError:
        Protocol = object


__all__ = ["RawTensorTypes", "ItemKeyType", "Tensor", "Dim", "get_raw_tensor_type", "GetModelFunc", "StepFunc"]


RawTensorTypes = Union[int, float, complex, numpy.number, numpy.ndarray, bool, str]
ItemKeyType = Union[RawTensorTypes, Tensor, slice, Sequence[Union[RawTensorTypes, Tensor, slice]]]


class GetModelFunc(Protocol):
    """get model func"""

    def __call__(self, *, epoch: int, step: int) -> rf.Module:
        ...


class StepFunc(Protocol):
    """step func"""

    def __call__(self, *, model: rf.Module, extern_data: TensorDict) -> None:
        ...


def get_raw_tensor_type() -> Type:
    """
    :return: the raw tensor type of the current selected backend, e.g. ``torch.Tensor`` or ``tf.Tensor``
    """
    if TYPE_CHECKING:
        import torch

        return torch.Tensor  # just as an example
    from ._backend import global_backend

    return global_backend.RawTensorType
